import os
import subprocess
import sys
from pathlib import Path
from argparse import ArgumentParser
import webbrowser
from contextlib import closing
import socket
import logging
import time
import platform
import tempfile
import shutil


def get_viewer_path():
    return str(Path(__file__).parent)


def ensure_dependencies():
    """Ensure npm dependencies are installed"""
    static_dir = os.path.join(get_viewer_path(), "static")
    node_modules = os.path.join(static_dir, "node_modules")

    if not os.path.exists(node_modules):
        print("First run: Installing Node.js dependencies...")
        print("Your node_modules path: ", node_modules)
        try:
            subprocess.run(["npm", "install"], cwd=static_dir, check=True)
            print("Dependencies installed successfully.")
        except subprocess.CalledProcessError as e:
            print(f"Error installing dependencies: {e}")
            sys.exit(1)
        except FileNotFoundError:
            print(
                "Error: Node.js is not installed. Please install Node.js to run the viewer."
            )
            sys.exit(1)


def _setup_logging(level):
    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s] %(message)s",
        level=level,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def check_node_installed():
    """Check if Node.js is installed and return version if found"""
    try:
        result = subprocess.run(
            ["node", "--version"], capture_output=True, text=True, check=True
        )
        return result.stdout.strip()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return None


def main():
    parser = ArgumentParser(description="Curator Viewer")
    parser.add_argument(
        "--host",
        default="127.0.0.1",
        help="Host to run the server on (default: localhost)",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=3000,
        help="Port to run the server on (default: 3000)",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Enables debug logging for more verbose output",
    )
    args = parser.parse_args()

    _setup_logging(logging.DEBUG if args.verbose else logging.INFO)

    # Check if Node.js is installed
    node_version = check_node_installed()
    if not node_version:
        print("\nNode.js is required but not found on your system.")
        print("Please install Node.js from https://nodejs.org/")
        print("\nAfter installing:")
        print("1. Close and reopen your terminal")
        print("2. Verify installation by running: node --version")
        print("3. Run curator-viewer again")
        sys.exit(1)

    ensure_dependencies()

    # Set environment variables for the Next.js server
    env = os.environ.copy()
    env["NODE_ENV"] = "production"
    env["HOST"] = args.host
    env["PORT"] = str(args.port)

    # Start the Next.js server
    viewer_path = get_viewer_path()
    static_dir = os.path.join(viewer_path, "static")
    server_file = os.path.join(viewer_path, "server.js")

    if not os.path.exists(os.path.join(static_dir, ".next")):
        print(
            "Error: Next.js build artifacts not found. The package may not be built correctly."
        )
        sys.exit(1)

    try:
        subprocess.run(
            ["node", server_file], cwd=viewer_path, env=env, check=True
        )
    except subprocess.CalledProcessError as e:
        print(f"Error starting Next.js server: {e}")
        sys.exit(1)
    except FileNotFoundError:
        print(
            "Error: Node.js is not installed. Please install Node.js to run the viewer."
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



import asyncio
import json
import logging
import os
from typing import Optional, Type, TypeVar

import aiofiles
from openai import AsyncOpenAI
from pydantic import BaseModel
from tqdm import tqdm

from bespokelabs.curator.dataset import Dataset
from bespokelabs.curator.prompter.prompt_formatter import PromptFormatter
from bespokelabs.curator.request_processor.base_request_processor import (
    BaseRequestProcessor,
    GenericRequest,
    GenericResponse,
    parse_response_message,
)
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop

T = TypeVar("T")
logger = logging.getLogger(__name__)


class OpenAIBatchRequestProcessor(BaseRequestProcessor):
    def __init__(
        self,
        batch_size: int = 1000,
        model: str = "gpt-4o-mini",
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        check_interval: int = 10,
        api_key: str = os.getenv("OPENAI_API_KEY"),
        url: str = "https://api.openai.com/v1/chat/completions",
    ):
        super().__init__(batch_size)
        self.url: str = url
        self.api_key: str = api_key
        self.check_interval: int = check_interval
        self.temperature: float = temperature
        self.top_p: float = top_p

    def get_rate_limits(self) -> dict:
        """
        Function to get rate limits for a given annotator. Not available via response headers, so
        the following is based on tier 5 limits on Nov 6th, 2024.

        These rate limits vary per model
        and are determined by your organization's usage tier. View the following:
        https://platform.openai.com/docs/guides/rate-limits/usage-tiers
        https://platform.openai.com/settings/organization/limits

        Args:
            model (str): The model for which to get the rate limits.
            request_url (str): The request URL for which to get the rate limits.

        Returns:
            tuple[int, int]: A tuple containing the maximum number of requests and tokens per minute.
        """
        model_tpd = {
            "gpt-3.5-turbo": 5_000_000_000,
            "gpt-3.5-turbo-0125": 5_000_000_000,
            "gpt-3.5-turbo-1106": 5_000_000_000,
            "gpt-3.5-turbo-16k": 5_000_000_000,
            "gpt-3.5-turbo-instruct": 200_000,
            "gpt-3.5-turbo-instruct-0914": 200_000,
            "gpt-4": 150_000_000,
            "gpt-4-0613": 150_000_000,
            "gpt-4-turbo": 300_000_000,
            "gpt-4o": 10_000_000_000,
            "gpt-4o-mini": 15_000_000_000,
        }

        if self.model not in model_tpd:
            tpd = 1_000_000_000
        else:
            tpd = model_tpd[self.model]

        logger.info(
            f"Automatically set max_tokens_per_day to {tpd}, model: {self.model} "
        )

        rate_limits = {"max_tokens_per_day": tpd}

        return rate_limits

    def create_api_specific_request(
        self, generic_request: GenericRequest
    ) -> dict:
        """
        Creates a API-specific request body from a generic request body.

        Using the api_parallel_processor, we can store whatever we want in the metadata. We will store both the row and the index.
        This is so we can later construct the new dataset row.

        Returns:
            dict: API specific request body
        """
        # NOTE(Ryan): We can have a shared place that creates the body (since it is the same for both online and batch).
        if generic_request.response_format:
            body = {
                "model": generic_request.model,
                "messages": generic_request.messages,
                "response_format": {
                    "type": "json_schema",
                    "json_schema": {
                        # TODO(ryan): not sure if this should be something else.
                        # TODO(ryan): also not sure if we should use strict: True
                        "name": "output_schema",
                        "schema": generic_request.response_format,
                    },
                },
            }
        else:
            body = {
                "model": generic_request.model,
                "messages": generic_request.messages,
            }

        if self.temperature is not None:
            body["temperature"] = self.temperature

        if self.top_p is not None:
            body["top_p"] = self.top_p

        request = {
            "custom_id": str(generic_request.original_row_idx),
            "method": "POST",
            "url": "/v1/chat/completions",
            "body": body,
        }

        return request

    async def asubmit_batch(self, batch_file: str) -> dict:
        async_client = AsyncOpenAI()
        # Create a list to store API-specific requests
        api_specific_requests = []

        async with aiofiles.open(batch_file, "r") as file:
            file_content = await file.read()
            for line in file_content.splitlines():
                request = GenericRequest.model_validate_json(line)
                api_specific_request = self.create_api_specific_request(request)
                api_specific_requests.append(json.dumps(api_specific_request))

        # Join requests with newlines and encode to bytes for upload
        file_content = "\n".join(api_specific_requests).encode()

        batch_file_upload = await async_client.files.create(
            file=file_content, purpose="batch"
        )

        logger.info(f"File uploaded: {batch_file_upload}")

        batch_object = await async_client.batches.create(
            input_file_id=batch_file_upload.id,
            endpoint="/v1/chat/completions",
            completion_window="24h",
            metadata={
                "request_file_name": batch_file
            },  # for downloading the batch to similarly named responses file
        )
        logger.info(
            f"Batch request submitted, received batch object: {batch_object}"
        )
        # Explicitly close the client. Otherwise we get something like
        # future: <Task finished name='Task-46' coro=<AsyncClient.aclose() done ... >>
        await async_client.close()

        return batch_object

    def run(
        self,
        dataset: Optional[Dataset],
        working_dir: str,
        parse_func_hash: str,
        prompt_formatter: PromptFormatter,
    ) -> Dataset:
        """
        Uses the API to completing the specific map by calling the LLM.

        Args:
            dataset (Dataset): Dataset that is being mapped over
            working_dir (str): Working directory to save files (requests.jsonl, responses.jsonl, dataset.arrow)
            parse_func_hash (str): Hash of the parse_func to be used as the dataset file name
            prompt_formatter (PromptFormatter): Prompt formatter to be used to format the prompt

        Returns:
            Dataset: Completed dataset
        """
        requests_files = self.create_request_files(
            dataset, working_dir, prompt_formatter
        )
        batch_objects_file = f"{working_dir}/batch_objects.jsonl"

        # TODO(Ryan): we should have an easy way to cancel all batches in batch_objects.jsonl if the user realized they made a mistake
        if os.path.exists(batch_objects_file):
            logger.warning(
                f"Batch objects file already exists, skipping batch submission and resuming: {batch_objects_file}"
            )
        else:
            # upload requests files and submit batches
            # asyncio gather preserves order
            async def submit_all_batches():
                tasks = [
                    self.asubmit_batch(requests_files[i])
                    for i in range(len(requests_files))
                ]
                return await asyncio.gather(*tasks)

            batch_objects = run_in_event_loop(submit_all_batches())

            with open(batch_objects_file, "w") as f:
                # NOTE(Ryan): we can also store the request_file_name in this object here, instead of in the metadata during batch submission. Can find a nice abstraction across other batch APIs (e.g. claude)
                for obj in batch_objects:
                    f.write(json.dumps(obj.model_dump(), default=str) + "\n")
            logger.info(f"Batch objects written to {batch_objects_file}")

        # TODO(Ryan): Actually do accounting for tokens, so rate limits enforced locally.
        # NOTE(Ryan): Although this isn't really practical since the limits are for an entire day and an entire organization. Maybe skip this and just recognize what a rate limit error for batching looks like (need to try this on a low tier account).
        # rate_limits = self.get_rate_limits()
        # tpd = rate_limits["max_tokens_per_day"]
        # token_encoding_name = get_token_encoding_name(self.model)

        # TODO(Ryan): based on the files that are downloaded, update completed_ids. If any are errors, try to resubmit (depending on error type).
        # TODO(Ryan): This creates responses_0.jsonl, responses_1.jsonl, etc. errors named same way? or errors_0.jsonl, errors_1.jsonl?
        # TODO(Ryan): retries, resubmits on lagging batches - need to study this a little closer
        # TODO(Ryan): likely can add some logic for smarter check_interval based on batch size and if the batch has started or not, fine to do a dumb ping for now
        # NOTE(Ryan): If we allow for multiple heterogeneous requests per dataset row, we will need to update this.
        total_requests = 1 if dataset is None else len(dataset)

        async def watch_batches():
            batch_watcher = BatchWatcher(
                working_dir, check_interval=self.check_interval
            )
            await batch_watcher.watch(
                prompt_formatter.response_format, total_requests
            )
            # Explicitly close the client. Otherwise we get something like
            # future: <Task finished name='Task-46' coro=<AsyncClient.aclose() done ... >>
            await batch_watcher.close_client()

        run_in_event_loop(watch_batches())

        dataset = self.create_dataset_files(
            working_dir, parse_func_hash, prompt_formatter
        )

        return dataset


class BatchWatcher:
    def __init__(self, working_dir: str, check_interval) -> None:
        """Initialize BatchWatcher with batch objects file and check interval.

        Args:
            batch_objects_file (str): Path to the batch objects JSON file.
            check_interval (int): Time interval (in seconds) to check batch status.
        """
        self.client = AsyncOpenAI()
        with open(f"{working_dir}/batch_objects.jsonl", "r") as f:
            self.batch_objects = [json.loads(line) for line in f]
        self.batch_ids = [obj["id"] for obj in self.batch_objects]
        self.batch_id_to_request_file_name = {
            obj["id"]: obj["metadata"]["request_file_name"]
            for obj in self.batch_objects
        }
        self.batches = []
        self.check_interval = check_interval
        self.working_dir = working_dir

    async def close_client(self):
        await self.client.close()

    async def check_batch_status(self, batch_id: str) -> tuple[str, str]:
        """Check the status of a batch by its ID.

        Args:
            batch_id (str): The ID of the batch to check.

        Returns:
            tuple[str, str]: The batch ID and its status.
        """
        batch = await self.client.batches.retrieve(batch_id)
        logger.info(
            f"Batch {batch_id} status: {batch.status} requests: {batch.request_counts.completed}/{batch.request_counts.failed}/{batch.request_counts.total} completed/failed/total"
        )
        return batch_id, batch

    async def watch(
        self,
        response_format: Optional[Type[BaseModel]],
        total_requests: int,
    ) -> None:
        """Monitor the status of batches until all are completed (includes successfully, failed, expired or cancelled)."""

        completed_batches = {}
        pbar = tqdm(
            total=total_requests,
            desc="Completed OpenAI requests in batches",
            unit="request",
        )

        while len(completed_batches) < len(self.batch_ids):
            pbar.n = 0
            status_tasks = []
            for batch_id in self.batch_ids:
                if batch_id not in completed_batches:
                    status_tasks.append(self.check_batch_status(batch_id))
                else:
                    pbar.n = (
                        pbar.n
                        + completed_batches[batch_id].request_counts.completed
                        + completed_batches[batch_id].request_counts.failed
                    )

            batches = await asyncio.gather(*status_tasks)
            newly_completed_batches = []
            for batch_id, batch in batches:
                if batch.status in [
                    "completed",
                    "failed",
                    "expired",
                    "cancelled",
                ]:
                    logger.info(
                        f"Batch {batch_id} processing finished with status: {batch.status}"
                    )
                    completed_batches[batch_id] = batch
                    newly_completed_batches.append(batch)

                pbar.n = (
                    pbar.n
                    + batch.request_counts.completed
                    + batch.request_counts.failed
                )

            pbar.refresh()

            tasks = [
                self.download_batch_to_generic_responses_file(
                    batch, response_format
                )
                for batch in newly_completed_batches
            ]
            await asyncio.gather(*tasks)

            if len(completed_batches) < len(self.batch_ids):
                logger.info(
                    f"Batches fully finished: {len(completed_batches)}/{len(self.batch_ids)} Requests completed: {pbar.n}/{total_requests}"
                )
                logger.info(f"Sleeping for {self.check_interval} seconds...")
                await asyncio.sleep(self.check_interval)

        pbar.close()
        self.batches = completed_batches.values()

    async def download_batch_to_generic_responses_file(
        self,
        batch,
        response_format: Optional[Type[BaseModel]],
    ) -> str:
        """Download the result of a completed batch to file.

        Args:
            batch: The batch object to download results from.

        Returns:
            str: Path to the downloaded result file.
        """
        if batch.status == "completed" and batch.output_file_id:
            file_content = await self.client.files.content(batch.output_file_id)
        elif batch.status == "failed" and batch.error_file_id:
            file_content = await self.client.files.content(batch.error_file_id)
        elif batch.status == "failed" and not batch.error_file_id:
            errors = "\n".join([str(error) for error in batch.errors.data])
            logger.warning(
                f"Batch {batch.id} failed\n" f"Batch errors: {errors}"
            )
            return None
        elif batch.status == "cancelled" or batch.status == "expired":
            logger.warning(f"Batch {batch.id} was cancelled or expired")
            return None

        # Naming is consistent with the request file (e.g. requests_0.jsonl -> responses_0.jsonl)
        request_file = self.batch_id_to_request_file_name[batch.id]
        request_file_idx = request_file.split("/")[-1].split("_", 1)[1]
        response_file = f"{self.working_dir}/responses_{request_file_idx}"

        generic_request_map = {}
        with open(request_file, "r") as f:
            for line in f:
                generic_request = GenericRequest.model_validate_json(line)
                generic_request_map[generic_request.original_row_idx] = (
                    generic_request
                )

        with open(response_file, "w") as f:
            for raw_response in file_content.text.splitlines():
                raw_response = json.loads(raw_response)
                generic_request = generic_request_map[
                    int(raw_response["custom_id"])
                ]

                generic_response = GenericResponse(
                    raw_response=raw_response,
                    raw_request=None,
                    generic_request=generic_request,
                )

                # TODO(Ryan): Add more specific error handling
                status_code = raw_response["response"]["status_code"]
                if status_code != 200:
                    logger.warning(
                        f"Request {generic_request} failed with status code {status_code}"
                    )
                    generic_response.response_errors = [
                        f"Request {generic_request} failed with status code {status_code}"
                    ]
                else:
                    # NOTE(Ryan): can we actually parse the response into a an OpenAI ChatCompletions object? Easier to access fields?
                    # TODO(Ryan): if you add token tokens to generic response
                    choices = raw_response["response"]["body"]["choices"]
                    # Assuming N = 1
                    response_message = choices[0]["message"]["content"]
                    response_message, response_errors = parse_response_message(
                        response_message, response_format
                    )
                    generic_response.response_message = response_message
                    generic_response.response_errors = response_errors
                f.write(
                    json.dumps(generic_response.model_dump(), default=str)
                    + "\n"
                )
        return response_file



import json
import logging
import os
import glob

import pandas as pd

from pydantic import BaseModel
from datasets import Dataset as HFDataset
from datasets.arrow_writer import ArrowWriter, SchemaInferenceError
from typing import Any, Dict, Iterable, Iterator, List, TypeVar

from bespokelabs.curator.prompter.prompt_formatter import PromptFormatter
from bespokelabs.curator.request_processor.generic_response import (
    GenericResponse,
)

T = TypeVar("T")


class Dataset:
    def __init__(
        self,
        iterable: Iterable[Dict[str, Any] | BaseModel] | None = None,
        working_dir: str | None = None,
        prompt_formatter: PromptFormatter | None = None,
    ):
        self.working_dir = working_dir
        self.prompt_formatter = prompt_formatter
        self.iterable = iterable

    def from_iterable(iterable: Iterable[Dict[str, Any] | BaseModel]):
        return Dataset(iterable=iterable)

    def from_working_dir(working_dir: str, prompt_formatter: PromptFormatter):
        return Dataset(
            working_dir=working_dir, prompt_formatter=prompt_formatter
        )

    def __iter__(self) -> Iterator[Dict[str, Any] | BaseModel]:
        if self.iterable is not None:
            yield from self.iterable
            return

        if self.working_dir:
            response_file = f"{self.working_dir}/responses.jsonl"

            for line in open(response_file, "r"):
                response = GenericResponse.model_validate_json(line)
                if self.prompt_formatter.response_format:
                    response.response = self.prompt_formatter.response_format(
                        **response.response
                    )
                if self.prompt_formatter.parse_func:
                    response = self.prompt_formatter.parse_func(
                        response.row, response.response
                    )
                else:
                    response = [response.response]

                if isinstance(response, list):
                    yield from response
                else:
                    yield response

    def to_list(self) -> List[Dict[str, Any] | BaseModel]:
        return list(self)

    def to_huggingface(self, in_memory: bool = False) -> None:
        """
        Returns a HuggingFace Dataset

        Args:
            in_memory (bool): Whether to load the dataset into memory

        Returns:
            Dataset: Completed dataset
        """

        total_responses_count = 0
        failed_responses_count = 0

        os.makedirs(self.working_dir, exist_ok=True)

        dataset_file = f"{self.working_dir}/dataset.arrow"
        responses_files = glob.glob(f"{self.working_dir}/responses_*.jsonl")
        if len(responses_files) == 0:
            raise ValueError(
                f"No responses files found in {self.working_dir}, can't construct dataset"
            )

        # Process all response files
        with ArrowWriter(path=dataset_file) as writer:
            for responses_file in responses_files:
                with open(responses_file, "r") as f_in:
                    for line in f_in:
                        total_responses_count += 1
                        response = GenericResponse.model_validate_json(line)
                        if self.prompt_formatter.response_format:
                            response.response = (
                                self.prompt_formatter.response_format(
                                    **response.response
                                )
                            )

                        if response is None:
                            failed_responses_count += 1
                            continue

                        if self.prompt_formatter.parse_func:
                            dataset_rows = self.prompt_formatter.parse_func(
                                response.row, response.response
                            )
                        else:
                            dataset_rows = [response.response]

                        for row in dataset_rows:
                            if isinstance(row, BaseModel):
                                row = row.model_dump()
                            writer.write(row)

            logging.info(
                f"Read {total_responses_count} responses, {failed_responses_count} failed"
            )
            logging.info("Finalizing writer")

            if failed_responses_count == total_responses_count:
                raise ValueError("All requests failed")

            # TODO(Ryan): Look at what this file looks like before finalize. What happens during finalize?
            try:
                writer.finalize()
            except SchemaInferenceError as e:
                raise ValueError(
                    "Arrow writer is complaining about the schema: likely all of your parsed rows were None and writer.write only wrote None objects."
                ) from e

        return HFDataset.from_file(dataset_file, in_memory=in_memory)



"""Database class for storing metadata for Bella runs."""

import os
import sqlite3


class MetadataDB:
    """Database class for storing Bella run metadata."""

    def __init__(self, db_path: str):
        self.db_path = db_path

    def store_metadata(self, metadata: dict):
        """Store metadata about a Bella run in the database.

        Args:
            metadata: Dictionary containing run metadata with keys:
                - timestamp: ISO format timestamp
                - dataset_hash: Unique hash of input dataset
                - prompt_func: Source code of prompt function
                - model_name: Name of model used
                - response_format: JSON schema of response format
                - run_hash: Unique hash identifying the run
                - batch_mode: Boolean indicating batch mode or online mode (True = batch, False = online)
        """
        os.makedirs(os.path.dirname(self.db_path), exist_ok=True)
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS runs (
                    run_hash TEXT PRIMARY KEY,
                    dataset_hash TEXT,
                    prompt_func TEXT,
                    model_name TEXT,
                    response_format TEXT,
                    batch_mode BOOLEAN,
                    created_time TEXT,
                    last_edited_time TEXT
                )
                """
            )

            # Check if run_hash exists
            cursor.execute(
                "SELECT run_hash FROM runs WHERE run_hash = ?",
                (metadata["run_hash"],),
            )
            existing_run = cursor.fetchone()

            if existing_run:
                # Update last_edited_time for existing entry
                cursor.execute(
                    """
                    UPDATE runs 
                    SET last_edited_time = ?
                    WHERE run_hash = ?
                    """,
                    (metadata["timestamp"], metadata["run_hash"]),
                )
            else:
                # Insert new entry
                cursor.execute(
                    """
                    INSERT INTO runs (
                        run_hash, dataset_hash, prompt_func, model_name, 
                        response_format, batch_mode, created_time, last_edited_time
                    ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                    """,
                    (
                        metadata["run_hash"],
                        metadata["dataset_hash"],
                        metadata["prompt_func"],
                        metadata["model_name"],
                        metadata["response_format"],
                        metadata["batch_mode"],
                        metadata["timestamp"],
                        "-",
                    ),
                )
            conn.commit()



import logging

logger = logging.getLogger("bespokelabs.curator")

handler = logging.StreamHandler()
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARNING)



"""Curator: Bespoke Labs Synthetic Data Generation Library."""

import inspect
import os
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, Optional, Type, TypeVar, Union

from datasets import Dataset
from pydantic import BaseModel
from xxhash import xxh64
import logging

from bespokelabs.curator.db import MetadataDB
from bespokelabs.curator.prompter.prompt_formatter import PromptFormatter
from bespokelabs.curator.request_processor.base_request_processor import (
    BaseRequestProcessor,
)
from bespokelabs.curator.request_processor.generic_request import GenericRequest
from bespokelabs.curator.request_processor.openai_batch_request_processor import (
    OpenAIBatchRequestProcessor,
)
from bespokelabs.curator.request_processor.openai_online_request_processor import (
    OpenAIOnlineRequestProcessor,
)

_CURATOR_DEFAULT_CACHE_DIR = "~/.cache/curator"
T = TypeVar("T")

logger = logging.getLogger(__name__)


class Prompter:
    """Interface for prompting LLMs."""

    def __init__(
        self,
        model_name: str,
        prompt_func: Callable[
            [Union[Dict[str, Any], BaseModel]], Dict[str, str]
        ],
        parse_func: Optional[
            Callable[
                [
                    Union[Dict[str, Any], BaseModel],
                    Union[Dict[str, Any], BaseModel],
                ],
                T,
            ]
        ] = None,
        response_format: Optional[Type[BaseModel]] = None,
        batch: bool = False,
        batch_size: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ):
        """Initialize a Prompter.

        Args:
            model_name (str): The name of the LLM to use
            prompt_func (Callable[[Dict[str, Any]], Union[str, List[Dict[str, Any]]]]): A function that takes a single row
                and returns either a string (assumed to be a user prompt) or messages list
            parse_func (Callable[[Dict[str, Any], Any], T]): A function that takes the input row and
                response object and returns the parsed output
            response_format (Optional[Type[BaseModel]]): A Pydantic model specifying the
                response format from the LLM.
            batch (bool): Whether to use batch processing
            batch_size (Optional[int]): The size of the batch to use, only used if batch is True
        """
        prompt_sig = inspect.signature(prompt_func)
        if len(prompt_sig.parameters) > 1:
            raise ValueError(
                f"prompt_func must take one argument or less, got {len(prompt_sig.parameters)}"
            )

        if parse_func is not None:
            parse_sig = inspect.signature(parse_func)
            if len(parse_sig.parameters) != 2:
                raise ValueError(
                    f"parse_func must take exactly 2 arguments, got {len(parse_sig.parameters)}"
                )

        self.prompt_formatter = PromptFormatter(
            model_name, prompt_func, parse_func, response_format
        )
        self.batch_mode = batch
        if batch:
            self._request_processor = OpenAIBatchRequestProcessor(
                model=model_name,
                batch_size=batch_size,
                temperature=temperature,
                top_p=top_p,
            )
        else:
            if batch_size is not None:
                logger.warning(
                    f"Prompter argument `batch_size` {batch_size} is ignored because `batch` is False"
                )
            self._request_processor = OpenAIOnlineRequestProcessor(
                model=model_name, temperature=temperature, top_p=top_p
            )

    def __call__(
        self, dataset: Optional[Iterable] = None, working_dir: str = None
    ) -> Dataset:
        """
        Run completions on a dataset.

        Args:
            dataset (Iterable): A dataset consisting of a list of items to apply completions
            working_dir (str): The working directory to save the requests.jsonl, responses.jsonl, and dataset.arrow files.
        """
        return self._completions(self._request_processor, dataset, working_dir)

    def _completions(
        self,
        request_processor: BaseRequestProcessor,
        dataset: Optional[Iterable] = None,
        working_dir: str = None,
    ) -> Dataset:
        """
        Apply structured completions in parallel to a dataset using specified model and
        prompts.

        Args:
            dataset (Iterable): A dataset consisting of a list of items to apply completions
            prompter (Prompter): A Prompter that contains the logic for formatting each
                item in the dataset
            working_dir (str): The working directory to save the requests.jsonl, responses.jsonl, and dataset.arrow files.

        Returns:
            Iterable: A list of structured outputs from the completions
        """
        # NOTE(Ryan): We convert from iterable to Dataset because Dataset has random access via row_idx
        if not isinstance(dataset, Dataset) and dataset is not None:
            dataset = Dataset.from_generator(dataset)

        if self is None:
            raise ValueError("Prompter must be provided")

        if working_dir is None:
            curator_cache_dir = os.environ.get(
                "CURATOR_CACHE_DIR",
                os.path.expanduser(_CURATOR_DEFAULT_CACHE_DIR),
            )
        else:
            curator_cache_dir = working_dir

        dataset_hash = (
            dataset._fingerprint
            if dataset is not None
            else xxh64("").hexdigest()
        )

        prompt_func_hash = _get_function_hash(self.prompt_formatter.prompt_func)

        # Used to name the dataset .arrow file, but not the cache directory name
        # Modifying `parse_func` creates a new dataset file from cached responses
        parse_func_hash = _get_function_hash(self.prompt_formatter.parse_func)

        fingerprint_str = "_".join(
            [
                str(dataset_hash),
                str(prompt_func_hash),
                str(self.prompt_formatter.model_name),
                str(
                    self.prompt_formatter.response_format.schema_json()
                    if self.prompt_formatter.response_format
                    else "text"
                ),
                str(self.batch_mode),
            ]
        )

        fingerprint = xxh64(fingerprint_str.encode("utf-8")).hexdigest()
        metadata_db_path = os.path.join(curator_cache_dir, "metadata.db")
        metadata_db = MetadataDB(metadata_db_path)

        # Get the source code of the prompt function
        prompt_func_source = inspect.getsource(
            self.prompt_formatter.prompt_func
        )
        if self.prompt_formatter.parse_func is not None:
            parse_func_source = inspect.getsource(
                self.prompt_formatter.parse_func
            )
        else:
            parse_func_source = ""

        metadata_dict = {
            "timestamp": datetime.now().isoformat(),
            "dataset_hash": dataset_hash,
            "prompt_func": prompt_func_source,
            "parse_func": parse_func_source,
            "model_name": self.prompt_formatter.model_name,
            "response_format": (
                self.prompt_formatter.response_format.schema_json()
                if self.prompt_formatter.response_format
                else "text"
            ),
            "run_hash": fingerprint,
            "batch_mode": self.batch_mode,
        }
        metadata_db.store_metadata(metadata_dict)

        dataset = request_processor.run(
            dataset=dataset,
            working_dir=os.path.join(curator_cache_dir, fingerprint),
            parse_func_hash=parse_func_hash,
            prompt_formatter=self.prompt_formatter,
        )

        return dataset


def _get_function_hash(func) -> str:
    """Get a hash of a function's source code."""
    if func is None:
        return xxh64("").hexdigest()

    return xxh64(inspect.getsource(func)).hexdigest()



import asyncio
import json
import logging
import os
import re
import time
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, Optional, Set, Tuple, TypeVar
import resource

import aiohttp
import requests
import tiktoken
from tqdm import tqdm

from bespokelabs.curator.dataset import Dataset
from bespokelabs.curator.prompter.prompter import PromptFormatter
from bespokelabs.curator.request_processor.base_request_processor import (
    BaseRequestProcessor,
    GenericRequest,
    GenericResponse,
    parse_response_message,
)
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop

T = TypeVar("T")
logger = logging.getLogger(__name__)


class OpenAIOnlineRequestProcessor(BaseRequestProcessor):
    def __init__(
        self,
        model: str = "gpt-4o-mini",
        api_key: str = os.getenv("OPENAI_API_KEY"),
        url: str = "https://api.openai.com/v1/chat/completions",
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
    ):
        super().__init__(batch_size=None)
        self.model: str = model
        self.url: str = url
        self.api_key: str = api_key
        self.temperature: float = temperature
        self.top_p: float = top_p

    def get_rate_limits(self) -> dict:
        """
        Function to get rate limits for a given annotator. Makes a single request to openAI API
        and gets the rate limits from the response headers. These rate limits vary per model
        and are determined by your organization's usage tier. View the following:
        https://platform.openai.com/docs/guides/rate-limits/usage-tiers
        https://platform.openai.com/settings/organization/limits

        Args:
            model (str): The model for which to get the rate limits.
            request_url (str): The request URL for which to get the rate limits.

        Returns:
            tuple[int, int]: A tuple containing the maximum number of requests and tokens per minute.
        """
        # Send a dummy request to get rate limit information
        response = requests.post(
            self.url,
            headers={"Authorization": f"Bearer {self.api_key}"},
            json={"model": self.model, "messages": []},
        )

        rpm = int(response.headers.get("x-ratelimit-limit-requests", 0))
        tpm = int(response.headers.get("x-ratelimit-limit-tokens", 0))

        if not rpm or not tpm:
            logger.warning(
                "Failed to get rate limits from OpenAI API, using default values"
            )
            rpm = 30_000
            tpm = 150_000_000

        logger.info(f"Automatically set max_requests_per_minute to {rpm}")
        logger.info(f"Automatically set max_tokens_per_minute to {tpm}")

        rate_limits = {
            "max_requests_per_minute": rpm,
            "max_tokens_per_minute": tpm,
        }

        return rate_limits

    def create_api_specific_request(
        self, generic_request: GenericRequest
    ) -> dict:
        """
        Creates a API-specific request body from a generic request body.

        Using the api_parallel_processor, we can store whatever we want in the metadata. We will store both the row and the index.
        This is so we can later construct the new dataset row.

        Returns:
            dict: API specific request body
        """
        request = {
            "model": generic_request.model,
            "messages": generic_request.messages,
        }
        if generic_request.response_format:
            request["response_format"] = {
                "type": "json_schema",
                "json_schema": {
                    "name": "output_schema",
                    "schema": generic_request.response_format,
                },
            }

        if self.temperature is not None:
            request["temperature"] = self.temperature

        if self.top_p is not None:
            request["top_p"] = self.top_p

        return request

    def run(
        self,
        dataset: Optional[Dataset],
        working_dir: str,
        parse_func_hash: str,
        prompt_formatter: PromptFormatter,
    ) -> Dataset:
        """
        Uses the API to completing the specific map by calling the LLM.

        Args:
            dataset (Dataset): Dataset that is being mapped over
            working_dir (str): Working directory to save files (requests.jsonl, responses.jsonl, dataset.arrow)
            parse_func_hash (str): Hash of the parse_func to be used as the dataset file name
            prompt_formatter (PromptFormatter): Prompt formatter to be used to format the prompt

        Returns:
            Dataset: Completed dataset
        """
        generic_requests_files = self.create_request_files(
            dataset, working_dir, prompt_formatter
        )
        generic_responses_files = [
            f"{working_dir}/responses_{i}.jsonl"
            for i in range(len(generic_requests_files))
        ]

        rate_limits = self.get_rate_limits()
        rpm = rate_limits["max_requests_per_minute"]
        tpm = rate_limits["max_tokens_per_minute"]

        token_encoding_name = get_token_encoding_name(
            prompt_formatter.model_name
        )

        # NOTE(Ryan): If you wanted to do this on batches, you could run a for loop here about request_files. Although I don't recommend it because you are waiting for straggler requests to finish for each batch.
        # NOTE(Ryan): And if you wanted to do batches in parallel, you would have to divide rpm and tpm by the number of parallel batches.
        # TODO(Ryan): Can we abstract retries from process_api_requests_from_file so you can use it even if you use liteLLM.
        for generic_requests_file, generic_responses_file in zip(
            generic_requests_files, generic_responses_files
        ):
            run_in_event_loop(
                self.process_generic_requests_from_file(
                    generic_requests_filepath=generic_requests_file,
                    save_filepath=generic_responses_file,
                    request_url=self.url,
                    max_requests_per_minute=rpm,
                    max_tokens_per_minute=tpm,
                    token_encoding_name=token_encoding_name,
                    max_attempts=5,
                    resume=True,  # detects existing jobs and resume from there
                )
            )

        dataset = self.create_dataset_files(
            working_dir, parse_func_hash, prompt_formatter
        )
        return dataset

    async def process_generic_requests_from_file(
        self,
        generic_requests_filepath: str,
        save_filepath: str,
        request_url: str,
        max_requests_per_minute: float,
        max_tokens_per_minute: float,
        token_encoding_name: str,
        max_attempts: int,
        resume: bool,
        resume_no_retry: bool = False,
    ) -> None:
        """Processes API requests in parallel, throttling to stay under rate limits."""

        # Increase the number of open file descriptors to avoid "Too many open files" errors
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        resource.setrlimit(
            resource.RLIMIT_NOFILE,
            (min(hard, 10 * max_requests_per_minute), hard),
        )

        # constants
        seconds_to_pause_after_rate_limit_error = 15
        seconds_to_sleep_each_loop = (
            0.001  # 1 ms limits max throughput to 1,000 requests per second
        )

        # infer API endpoint and construct request header
        api_endpoint = api_endpoint_from_url(self.url)
        request_header = {"Authorization": f"Bearer {self.api_key}"}
        # use api-key header for Azure deployments
        if "/deployments" in self.url:
            request_header = {"api-key": f"{self.api_key}"}

        # initialize trackers
        queue_of_requests_to_retry = asyncio.Queue()
        task_id_generator = (
            task_id_generator_function()
        )  # generates integer IDs of 0, 1, 2, ...
        status_tracker = (
            StatusTracker()
        )  # single instance to track a collection of variables
        next_request = None  # variable to hold the next request to call

        # initialize available capacity counts
        available_request_capacity = max_requests_per_minute
        available_token_capacity = max_tokens_per_minute
        last_update_time = time.time()

        # initialize flags
        file_not_finished = True  # after file is empty, we'll skip reading it
        logger.debug(f"Initialization complete.")

        completed_request_ids: Set[int] = set()
        if os.path.exists(save_filepath):
            if resume:
                # save all successfully completed requests to a temporary file, then overwrite the original file with the temporary file
                logger.debug(
                    f"Resuming progress from existing file: {save_filepath}"
                )
                logger.debug(
                    f"Removing all failed requests from {save_filepath} so they can be retried"
                )
                temp_filepath = f"{save_filepath}.temp"
                num_previously_failed_requests = 0
                with open(save_filepath, "r") as input_file, open(
                    temp_filepath, "w"
                ) as output_file:
                    for line in input_file:
                        response = GenericResponse.model_validate_json(line)
                        if response.response_errors:
                            # this means that the request failed and we have a list of errors
                            logger.debug(
                                f"Request {response.generic_request.original_row_idx} previously failed due to errors: {response.response_errors}, removing from output and will retry"
                            )
                            num_previously_failed_requests += 1
                        else:
                            completed_request_ids.add(
                                response.generic_request.original_row_idx
                            )
                            output_file.write(line)
                logger.info(
                    f"Found {len(completed_request_ids)} completed requests and {num_previously_failed_requests} previously failed requests"
                )
                logger.info(
                    "Failed requests and remaining requests will now be processed."
                )
                os.replace(temp_filepath, save_filepath)
            elif resume_no_retry:
                logger.warning(
                    f"Resuming progress from existing file: {save_filepath}, without retrying failed requests"
                )
                num_previously_failed_requests = 0
                with open(save_filepath, "r") as input_file, open(
                    temp_filepath, "w"
                ) as output_file:
                    for line in tqdm(
                        input_file, desc="Processing existing requests"
                    ):
                        data = json.loads(line)
                        if isinstance(data[1], list):
                            # this means that the request failed and we have a list of errors
                            logger.debug(
                                f"Request {data[2].get('request_idx')} previously failed due to errors: {data[1]}, will NOT retry"
                            )
                            num_previously_failed_requests += 1
                        completed_request_ids.add(data[2].get("request_idx"))
                logger.info(
                    f"Found {len(completed_request_ids)} total requests and {num_previously_failed_requests} previously failed requests"
                )
                logger.info("Remaining requests will now be processed.")
            else:
                user_input = input(
                    f"File {save_filepath} already exists.\nTo resume if there are remaining requests without responses, run with --resume flag.\nOverwrite? (Y/n): "
                )
                if user_input.lower() != "y" and user_input.lower() != "":
                    logger.info("Aborting operation.")
                    return

        # initialize file reading
        with open(generic_requests_filepath) as file:
            # `requests` will provide requests one at a time
            generic_requests = file.__iter__()
            logger.debug(f"File opened. Entering main loop")

            # Count total number of requests
            total_requests = sum(1 for _ in open(generic_requests_filepath))
            if total_requests == len(completed_request_ids):
                logger.debug(
                    "All requests have already been completed so will just reuse cache."
                )
                return

            # Create progress bar
            pbar = tqdm(
                total=total_requests,
                desc="Processing parallel requests to OpenAI",
            )

            connector = aiohttp.TCPConnector(limit=10 * max_requests_per_minute)
            async with aiohttp.ClientSession(
                connector=connector
            ) as session:  # Initialize ClientSession here
                while True:
                    # get next request (if one is not already waiting for capacity)
                    if next_request is None:
                        if not queue_of_requests_to_retry.empty():
                            next_request = (
                                queue_of_requests_to_retry.get_nowait()
                            )
                            logger.debug(
                                f"Retrying request {next_request.task_id}: {next_request}"
                            )
                        elif file_not_finished:
                            try:
                                # get new generic request
                                generic_request_json = json.loads(
                                    next(generic_requests)
                                )
                                generic_request = GenericRequest.model_validate(
                                    generic_request_json
                                )
                                request_idx = generic_request.original_row_idx

                                # Skip requests we already have responses for
                                if (
                                    resume
                                    and request_idx in completed_request_ids
                                ):
                                    logger.debug(
                                        f"Skipping already completed request {request_idx}"
                                    )
                                    status_tracker.num_tasks_already_completed += (
                                        1
                                    )
                                    continue

                                # Create API-specific request
                                api_specific_request_json = (
                                    self.create_api_specific_request(
                                        generic_request
                                    )
                                )
                                next_request = APIRequest(
                                    task_id=next(task_id_generator),
                                    api_specific_request_json=api_specific_request_json,
                                    generic_request=generic_request,
                                    token_consumption=num_tokens_consumed_from_request(
                                        api_specific_request_json,
                                        api_endpoint,
                                        token_encoding_name,
                                    ),
                                    attempts_left=max_attempts,
                                )
                                status_tracker.num_tasks_started += 1
                                status_tracker.num_tasks_in_progress += 1
                                logger.debug(
                                    f"Reading request {next_request.task_id}: {next_request}"
                                )
                            except StopIteration:
                                # if file runs out, set flag to stop reading it
                                logger.debug("Read file exhausted")
                                file_not_finished = False

                    # update available capacity
                    current_time = time.time()
                    seconds_since_update = current_time - last_update_time
                    available_request_capacity = min(
                        available_request_capacity
                        + max_requests_per_minute * seconds_since_update / 60.0,
                        max_requests_per_minute,
                    )
                    available_token_capacity = min(
                        available_token_capacity
                        + max_tokens_per_minute * seconds_since_update / 60.0,
                        max_tokens_per_minute,
                    )
                    last_update_time = current_time

                    # if enough capacity available, call API
                    if next_request:
                        next_request_tokens = next_request.token_consumption
                        if (
                            available_request_capacity >= 1
                            and available_token_capacity >= next_request_tokens
                        ):
                            # update counters
                            available_request_capacity -= 1
                            available_token_capacity -= next_request_tokens
                            next_request.attempts_left -= 1

                            # call API
                            asyncio.create_task(
                                next_request.call_api(
                                    session=session,
                                    request_url=request_url,
                                    request_header=request_header,
                                    retry_queue=queue_of_requests_to_retry,
                                    save_filepath=save_filepath,
                                    status_tracker=status_tracker,
                                ),
                            )
                            next_request = None  # reset next_request to empty
                        else:
                            logger.debug(
                                f"Not Enough Capacity: Request tokens: {next_request_tokens}, Available request capacity: {available_request_capacity}, Available token capacity: {available_token_capacity}"
                            )

                    # Update progress bar when a task is completed
                    total_completed = (
                        status_tracker.num_tasks_succeeded
                        + status_tracker.num_tasks_failed
                        + status_tracker.num_tasks_already_completed
                    )
                    if total_completed > pbar.n:
                        pbar.update(total_completed - pbar.n)

                    # if all tasks are finished, break
                    if status_tracker.num_tasks_in_progress == 0:
                        break

                    # main loop sleeps briefly so concurrent tasks can run
                    await asyncio.sleep(seconds_to_sleep_each_loop)

                    # if a rate limit error was hit recently, pause to cool down
                    seconds_since_rate_limit_error = (
                        time.time()
                        - status_tracker.time_of_last_rate_limit_error
                    )
                    if (
                        seconds_since_rate_limit_error
                        < seconds_to_pause_after_rate_limit_error
                    ):
                        remaining_seconds_to_pause = (
                            seconds_to_pause_after_rate_limit_error
                            - seconds_since_rate_limit_error
                        )
                        await asyncio.sleep(remaining_seconds_to_pause)
                        # ^e.g., if pause is 15 seconds and final limit was hit 5 seconds ago
                        logger.warn(
                            f"Pausing to cool down until {time.ctime(status_tracker.time_of_last_rate_limit_error + seconds_to_pause_after_rate_limit_error)}"
                        )

            # Close the progress bar
            pbar.close()

            # after finishing, log final status
            logger.info(
                f"""Parallel processing complete. Results saved to {save_filepath}"""
            )

            logger.info(f"Status tracker: {status_tracker}")

            if status_tracker.num_tasks_failed > 0:
                logger.warning(
                    f"{status_tracker.num_tasks_failed} / {status_tracker.num_tasks_started} requests failed. Errors logged to {save_filepath}."
                )
            if status_tracker.num_rate_limit_errors > 0:
                logger.warning(
                    f"{status_tracker.num_rate_limit_errors} rate limit errors received. Consider running at a lower rate."
                )


@dataclass
class StatusTracker:
    """Stores metadata about the script's progress. Only one instance is created."""

    num_tasks_already_completed: int = 0
    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0  # script ends when this reaches 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0  # excluding rate limit errors, counted above
    num_other_errors: int = 0
    time_of_last_rate_limit_error: int = (
        0  # used to cool off after hitting rate limits
    )


@dataclass
class APIRequest:
    """Stores an API request's inputs, outputs, and other metadata. Contains a method to make an API call."""

    task_id: int
    generic_request: GenericRequest
    api_specific_request_json: dict
    token_consumption: int
    attempts_left: int
    result: list = field(default_factory=list)

    async def call_api(
        self,
        session: aiohttp.ClientSession,
        request_url: str,
        request_header: dict,
        retry_queue: asyncio.Queue,
        save_filepath: str,
        status_tracker: StatusTracker,
    ) -> None:
        """Calls the OpenAI API and saves results."""
        logger.debug(f"Starting request #{self.task_id}")
        error = None
        try:
            async with session.post(
                url=request_url,
                headers=request_header,
                json=self.api_specific_request_json,
            ) as response:
                response = await response.json()
            if "error" in response:
                logger.warning(
                    f"Request {self.task_id} failed with error {response['error']}"
                )
                status_tracker.num_api_errors += 1
                error = response
                if "rate limit" in response["error"].get("message", "").lower():
                    status_tracker.time_of_last_rate_limit_error = time.time()
                    status_tracker.num_rate_limit_errors += 1
                    status_tracker.num_api_errors -= (
                        1  # rate limit errors are counted separately
                    )

        except (
            Exception
        ) as e:  # catching naked exceptions is bad practice, but in this case we'll log & save them
            logger.warning(
                f"Request {self.task_id} failed with Exception {e}, attempts left {self.attempts_left}"
            )
            status_tracker.num_other_errors += 1
            error = e
        if error:
            self.result.append(error)
            if self.attempts_left:
                retry_queue.put_nowait(self)
            else:
                generic_response = GenericResponse(
                    response_message=None,
                    response_errors=[str(e) for e in self.result],
                    raw_request=self.api_specific_request_json,
                    raw_response=None,
                    generic_request=self.generic_request,
                )
                append_generic_response(generic_response, save_filepath)
                status_tracker.num_tasks_in_progress -= 1
                status_tracker.num_tasks_failed += 1
                logger.error(
                    f"Request {self.api_specific_request_json} failed after all attempts."
                    f"Saved errors {self.result} to {save_filepath}"
                )
        else:
            response_message = response["choices"][0]["message"]["content"]
            response_message, response_errors = parse_response_message(
                response_message, self.generic_request.response_format
            )
            generic_response = GenericResponse(
                response_message=response_message,
                response_errors=response_errors,
                raw_request=self.api_specific_request_json,
                raw_response=response,
                generic_request=self.generic_request,
            )
            append_generic_response(generic_response, save_filepath)
            status_tracker.num_tasks_in_progress -= 1
            status_tracker.num_tasks_succeeded += 1
            logger.debug(f"Request {self.task_id} saved to {save_filepath}")


def get_token_encoding_name(model: str) -> str:
    """Get the token encoding name for a given model."""
    if "gpt" in model:
        return tiktoken.encoding_for_model(model).name
    else:
        logger.warning(
            f'Token encoding name for model "{model}" not implemented, using cl100k_base for token counting'
        )
        return "cl100k_base"


def get_rate_limits(
    model: str, request_url: str, api_key: str
) -> Tuple[int, int]:
    """
    Function to get rate limits for a given annotator. Makes a single request to openAI API
    and gets the rate limits from the response headers. These rate limits vary per model
    and are determined by your organization's usage tier. View the following:
    https://platform.openai.com/docs/guides/rate-limits/usage-tiers
    https://platform.openai.com/settings/organization/limits

    Args:
        model (str): The model for which to get the rate limits.
        request_url (str): The request URL for which to get the rate limits.

    Returns:
        Tuple[int, int]: The maximum number of requests and tokens per minute.
    """
    if "api.openai.com" in request_url:
        # Send a dummy request to get rate limit information
        response = requests.post(
            request_url,
            headers={"Authorization": f"Bearer {api_key}"},
            json={"model": model, "messages": []},
        )
        # Extract rate limit information from headers
        max_requests = int(
            response.headers.get("x-ratelimit-limit-requests", 30_000)
        )
        max_tokens = int(
            response.headers.get("x-ratelimit-limit-tokens", 150_000_000)
        )
    elif "api.sambanova.ai" in request_url:
        # Send a dummy request to get rate limit information
        max_requests = 50
        max_tokens = 100_000_000
    else:
        raise NotImplementedError(
            f'Rate limits for API endpoint "{request_url}" not implemented'
        )

    return max_requests, max_tokens


def get_api_key(request_url: str) -> str:
    """Get the API key for a given request URL."""
    if "api.openai.com" in request_url:
        return os.getenv("OPENAI_API_KEY")
    elif "api.sambanova.ai" in request_url:
        return os.getenv("SAMBANOVA_API_KEY")
    else:
        raise NotImplementedError(
            f'Default API key environment variable for API endpoint "{request_url}" not implemented'
        )


def api_endpoint_from_url(request_url: str) -> str:
    """Extract the API endpoint from the request URL.
    This is used to determine the number of tokens consumed by the request.
    """

    # OpenAI API
    match = re.search("^https://[^/]+/v\\d+/(.+)$", request_url)
    if match:
        return match[1]

    # for Azure OpenAI deployment urls
    match = re.search(
        r"^https://[^/]+/openai/deployments/[^/]+/(.+?)(\?|$)", request_url
    )
    if match:
        return match[1]

    # Catch all for other API endpoints using OpenAI OpenAPI format
    if "chat/completions" in request_url:
        return "chat/completions"
    elif "completions" in request_url:
        return "completions"
    else:
        raise NotImplementedError(
            f'API endpoint "{request_url}" not implemented in this script'
        )


def append_generic_response(data: GenericResponse, filename: str) -> None:
    """Append a json payload to the end of a jsonl file."""
    json_string = json.dumps(data.model_dump(), default=str)
    with open(filename, "a") as f:
        f.write(json_string + "\n")


def num_tokens_consumed_from_request(
    api_specific_request_json: dict,
    api_endpoint: str,
    token_encoding_name: str,
):
    """Count the number of tokens in the request. Only supports completion and embedding requests."""
    encoding = tiktoken.get_encoding(token_encoding_name)
    # if completions request, tokens = prompt + n * max_tokens
    if api_endpoint.endswith("completions"):
        max_tokens = api_specific_request_json.get("max_tokens", 15)
        n = api_specific_request_json.get("n", 1)
        completion_tokens = n * max_tokens

        # chat completions
        if api_endpoint.startswith("chat/"):
            num_tokens = 0
            for message in api_specific_request_json["messages"]:
                num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
                for key, value in message.items():
                    try:
                        num_tokens += len(encoding.encode(str(value)))
                    except TypeError:
                        logger.warning(
                            f"Failed to encode value {value} with tiktoken to count tokens. Instead assuming a token for every 4 characters."
                        )
                        num_tokens += len(str(value)) // 4
                    if key == "name":  # if there's a name, the role is omitted
                        num_tokens -= (
                            1  # role is always required and always 1 token
                        )
            num_tokens += 2  # every reply is primed with <im_start>assistant
            return num_tokens + completion_tokens
        # normal completions
        else:
            prompt = api_specific_request_json["prompt"]
            if isinstance(prompt, str):  # single prompt
                prompt_tokens = len(encoding.encode(prompt))
                num_tokens = prompt_tokens + completion_tokens
                return num_tokens
            elif isinstance(prompt, list):  # multiple prompts
                prompt_tokens = sum([len(encoding.encode(p)) for p in prompt])
                num_tokens = prompt_tokens + completion_tokens * len(prompt)
                return num_tokens
            else:
                raise TypeError(
                    'Expecting either string or list of strings for "prompt" field in completion request'
                )
    # if embeddings request, tokens = input tokens
    elif api_endpoint == "embeddings":
        input = api_specific_request_json["input"]
        if isinstance(input, str):  # single input
            num_tokens = len(encoding.encode(input))
            return num_tokens
        elif isinstance(input, list):  # multiple inputs
            num_tokens = sum([len(encoding.encode(i)) for i in input])
            return num_tokens
        else:
            raise TypeError(
                'Expecting either string or list of strings for "inputs" field in embedding request'
            )
    # more logic needed to support other API calls (e.g., edits, inserts, DALL-E)
    else:
        raise NotImplementedError(
            f'API endpoint "{api_endpoint}" not implemented in this script'
        )


def task_id_generator_function():
    """Generate integers 0, 1, 2, and so on."""
    task_id = 0
    while True:
        yield task_id
        task_id += 1



import inspect
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union

from pydantic import BaseModel

from bespokelabs.curator.request_processor.generic_request import GenericRequest

T = TypeVar("T")


class PromptFormatter:
    model_name: str
    prompt_func: Callable[[Union[Dict[str, Any], BaseModel]], Dict[str, str]]
    parse_func: Optional[
        Callable[
            [
                Union[Dict[str, Any], BaseModel],
                Union[Dict[str, Any], BaseModel],
            ],
            T,
        ]
    ] = None
    response_format: Optional[Type[BaseModel]] = None

    def __init__(
        self,
        model_name: str,
        prompt_func: Callable[
            [Union[Dict[str, Any], BaseModel]], Dict[str, str]
        ],
        parse_func: Optional[
            Callable[
                [
                    Union[Dict[str, Any], BaseModel],
                    Union[Dict[str, Any], BaseModel],
                ],
                T,
            ]
        ] = None,
        response_format: Optional[Type[BaseModel]] = None,
    ):
        self.model_name = model_name
        self.prompt_func = prompt_func
        self.parse_func = parse_func
        self.response_format = response_format

    def create_generic_request(
        self, row: Dict[str, Any] | BaseModel, idx: int
    ) -> GenericRequest:
        """Format the request object based off Prompter attributes."""
        sig = inspect.signature(self.prompt_func)
        if len(sig.parameters) == 0:
            prompts = self.prompt_func()
        elif len(sig.parameters) == 1:
            prompts = self.prompt_func(row)
        else:
            raise ValueError(
                f"Prompting function {self.prompt_func} must have 0 or 1 arguments."
            )

        if isinstance(prompts, str):
            messages = [{"role": "user", "content": prompts}]
        else:
            # TODO(Ryan): Add validation here
            messages = prompts

        # Convert BaseModel to dict for serialization
        if isinstance(row, BaseModel):
            row = row.model_dump()

        return GenericRequest(
            model=self.model_name,
            messages=messages,
            original_row=row,
            original_row_idx=idx,
            response_format=(
                self.response_format.model_json_schema()
                if self.response_format
                else None
            ),
        )



from typing import Any, Dict, List, Optional, Type
from pydantic import BaseModel

"""A generic request model for LLM API requests.

Attributes:
    model: The name/identifier of the LLM model to use.
    messages: List of message dictionaries forming the conversation history.
    response_format: Optional json schema of a Pydantic model class for structured output validation.
        None indicates non-structured (str) output is expected.
    original_row: The source data being processed. The original row as a dictionary.
    original_row_idx: The index of the original row in the dataset.
"""


class GenericRequest(BaseModel):
    model: str
    messages: List[Dict[str, Any]]
    response_format: Dict[str, Any] | None = None
    original_row: Dict[str, Any]
    original_row_idx: int



from typing import Any, Dict, List, Optional
from pydantic import BaseModel
from .generic_request import GenericRequest

"""A generic response model for LLM API requests.

Attributes:
    response_message: The main response content. Can be:
        - None when there are errors
        - str for non-structured output
        - Dict[str, Any] for structured output
    response_errors: List of error messages. None when there are no errors.
    raw_response: The raw response data from the API.
    raw_request: The raw request data. Will be None for BatchAPI requests.
    generic_request: The associated GenericRequest object.
"""


class GenericResponse(BaseModel):
    response_message: Optional[Dict[str, Any]] | str = None
    response_errors: Optional[List[str]] = None
    raw_response: Optional[Dict[str, Any]]
    raw_request: Optional[Dict[str, Any]] = None
    generic_request: GenericRequest



import asyncio
import glob
import json
import logging
import os
from abc import ABC, abstractmethod
from math import ceil
from typing import Optional

import aiofiles
import pyarrow
from datasets import Dataset
from datasets.arrow_writer import ArrowWriter
from pydantic import BaseModel, ValidationError

from bespokelabs.curator.prompter.prompt_formatter import PromptFormatter
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop
from bespokelabs.curator.request_processor.generic_request import GenericRequest
from bespokelabs.curator.request_processor.generic_response import (
    GenericResponse,
)

logger = logging.getLogger(__name__)


class BaseRequestProcessor(ABC):
    """
    Base class for all request processors.
    """

    def __init__(self, batch_size: Optional[int] = None):
        self.batch_size = batch_size

    @abstractmethod
    def get_rate_limits(self) -> dict:
        """
        Returns the rate limits for the API.

        Returns:
            dict: A dictionary containing the rate limit information.
        """
        pass

    @abstractmethod
    def create_api_specific_request(
        self, generic_request: GenericRequest
    ) -> dict:
        """
        Creates a API-specific request body from a GenericRequest.

        Must use request_body.metadata.dataset_row or request_body.metadata.dataset_row_idx

        Note that the generic request body has both the original dataset row or the index of the row in the original dataset.
        What you pass on depends on what is works with the API. For example, OpenAI and Anthropic offline batch API allows you to pass a custom_id (can be index).
        For online, the api_parallel_processor can store the original dataset row in the metadata.

        Returns:
            dict: API specific request body
        """
        pass

    @abstractmethod
    def run(
        self,
        dataset: Dataset,
        working_dir: str,
        parse_func_hash: str,
        prompt_formatter: PromptFormatter,
    ) -> Dataset:
        """
        Uses the API to completing the specific map by calling the LLM.

        Args:
            dataset (Dataset): Dataset that is being mapped over
            working_dir (str): Working directory to save files (requests.jsonl, responses.jsonl, dataset.arrow)

        Returns:
            Dataset: Completed dataset
        """
        pass

    def create_request_files(
        self,
        dataset: Optional[Dataset],
        working_dir: str,
        prompt_formatter: PromptFormatter,
    ) -> list[str]:
        """
        Creates a request file if they don't already exist or use existing.

        Args:
            dataset (Dataset): The dataset to be processed.
            working_dir (str): The directory where request files will be saved.

        Returns:
            list[str]: Paths to the request files that were created.
        """
        os.makedirs(working_dir, exist_ok=True)
        requests_files = glob.glob(f"{working_dir}/requests_*.jsonl")

        # By default use existing requests in working_dir
        if len(requests_files) > 0:
            logger.info(
                f"Using existing requests in {working_dir} by default. Found {len(requests_files)} request files."
                f"If this is not what you want, delete the directory or specify a new one and re-run."
            )
            # count existing jobs in file and print first job
            with open(requests_files[0], "r") as f:
                # Count lines and store first job
                first_job = None
                num_jobs = 0
                for i, line in enumerate(f):
                    if i == 0:
                        first_job = json.loads(line)
                    num_jobs = i + 1

                if num_jobs > 0:
                    logger.info(
                        f"There are {num_jobs} existing requests in {requests_files[0]}"
                    )
                    logger.info(
                        f"Example request in {requests_files[0]}:\n{json.dumps(first_job, default=str, indent=2)}"
                    )
                    return requests_files

        # Create new requests file
        requests_file = f"{working_dir}/requests_0.jsonl"
        requests_files = [requests_file]

        if dataset is None:
            with open(requests_file, "w") as f:
                generic_request = prompt_formatter.create_generic_request(
                    dict(), 0
                )
                f.write(
                    json.dumps(generic_request.model_dump(), default=str) + "\n"
                )
            return requests_files

        if self.batch_size:
            num_batches = ceil(len(dataset) / self.batch_size)
            requests_files = [
                f"{working_dir}/requests_{i}.jsonl" for i in range(num_batches)
            ]

            async def create_all_request_files():
                tasks = [
                    self.acreate_request_file(
                        dataset,
                        prompt_formatter,
                        requests_files[i],
                        start_idx=i * self.batch_size,
                    )
                    for i in range(num_batches)
                ]
                await asyncio.gather(*tasks)

            run_in_event_loop(create_all_request_files())
        else:
            run_in_event_loop(
                self.acreate_request_file(
                    dataset, prompt_formatter, requests_file
                )
            )

        return requests_files

    # NOTE(Ryan): Instead of doing this, just iterate over iterable and keep counter and change filename when hit batch_size, this will be slower but this whole thing is dominated by llm calls anyways
    async def acreate_request_file(
        self,
        dataset: Dataset,
        prompt_formatter: PromptFormatter,
        request_file: str,
        start_idx: int = 0,
    ) -> str:
        if self.batch_size is not None:
            end_idx = min(start_idx + self.batch_size, len(dataset))
            dataset = dataset.select(range(start_idx, end_idx))
        else:
            end_idx = len(dataset)

        # NOTE(Ryan): For loops only for IterableDataset which allows for _very_ large datasets, when start_idx and batch_size are not specified
        async with aiofiles.open(request_file, "w") as f:
            for idx, dataset_row in enumerate(dataset):
                dataset_row_idx = idx + start_idx
                # Get the generic request from the map function
                request = prompt_formatter.create_generic_request(
                    dataset_row, dataset_row_idx
                )
                await f.write(
                    json.dumps(request.model_dump(), default=str) + "\n"
                )
        logger.info(f"Wrote {end_idx - start_idx} requests to {request_file}.")

    def create_dataset_files(
        self,
        working_dir: str,
        parse_func_hash: str,
        prompt_formatter: PromptFormatter,
    ) -> None:
        """
        Creates the request files if they don't already exist or use existing.
        A single request file (requests_0.jsonl) or multiple request files
        (requests_0.jsonl, requests_1.jsonl, etc.) are created depending on
        batch_size.

        Args:
            dataset (Dataset): The dataset to be processed.
            working_dir (str): The directory where request files will be saved.
            prompt_formatter (PromptFormatter): The prompt formatter to use for parsing the responses.

        Returns:
            Dataset: Completed dataset
        """
        total_responses_count = 0
        failed_responses_count = 0

        responses_files = glob.glob(f"{working_dir}/responses_*.jsonl")
        if len(responses_files) == 0:
            raise ValueError(f"No responses files found in {working_dir}")
        dataset_file = f"{working_dir}/{parse_func_hash}.arrow"
        if os.path.exists(dataset_file):
            logger.info(f"Using existing dataset file {dataset_file}")
            successful_dataset_cache_load = False
            try:
                output_dataset = Dataset.from_file(dataset_file)
                successful_dataset_cache_load = True
            except pyarrow.lib.ArrowInvalid as e:
                os.remove(dataset_file)
                logger.warning(
                    f"Failed to load dataset from {dataset_file}, "
                    "which was likely corrupted by a failed previous run. "
                    "Deleted file and attempting to regenerate dataset from cached LLM responses."
                )

            if successful_dataset_cache_load:
                return output_dataset

        error_help = (
            f"Please check your `parse_func` is returning a valid row (dict) "
            "or list of rows (list of dicts) and re-run. "
            "Dataset will be regenerated from cached LLM responses."
        )

        # Process all response files
        with ArrowWriter(path=dataset_file) as writer:
            for responses_file in responses_files:
                with open(responses_file, "r") as f_in:
                    for generic_response_string in f_in:
                        total_responses_count += 1
                        response = GenericResponse.model_validate_json(
                            generic_response_string
                        )

                        # response.response_errors is not None IFF response.response_message is None
                        if response.response_errors is not None:
                            failed_responses_count += 1
                            continue

                        if prompt_formatter.response_format:
                            # Response message is a string, which is converted to a dict
                            # The dict is then used to construct the response_format Pydantic model
                            try:
                                response.response_message = (
                                    prompt_formatter.response_format(
                                        **response.response_message
                                    )
                                )
                            except ValidationError as e:
                                schema_str = json.dumps(
                                    prompt_formatter.response_format.model_json_schema(),
                                    indent=2,
                                )
                                warning_msg = (
                                    f"Pydantic failed to parse response message {response.response_message} with `response_format` {schema_str}."
                                    f"The model likely returned a JSON that does not match the schema of the `response_format`. Will skip this response."
                                )
                                logger.warning(warning_msg)
                                failed_responses_count += 1
                                continue

                        # parse_func can return a single row or a list of rows
                        if prompt_formatter.parse_func:
                            try:
                                dataset_rows = prompt_formatter.parse_func(
                                    response.generic_request.original_row,
                                    response.response_message,
                                )
                            except Exception as e:
                                logger.error(
                                    f"Exception raised in your `parse_func`. {error_help}"
                                )
                                os.remove(dataset_file)
                                raise e
                            if not isinstance(dataset_rows, list):
                                dataset_rows = [dataset_rows]
                        else:
                            dataset_rows = [
                                {"response": response.response_message}
                            ]

                        for row in dataset_rows:
                            if isinstance(row, BaseModel):
                                row = row.model_dump()

                            if not isinstance(row, dict):
                                os.remove(dataset_file)
                                raise ValueError(
                                    f"Got invalid row {row} of type {type(row)} from `parse_func`. "
                                    f"This should be type <class 'dict'>. {error_help}"
                                )
                            if not row:
                                os.remove(dataset_file)
                                raise ValueError(
                                    f"Got empty row {row} from `parse_func`. {error_help}"
                                )

                            writer.write(row)

            logger.info(
                f"Read {total_responses_count} responses, {failed_responses_count} failed"
            )
            if failed_responses_count == total_responses_count:
                os.remove(dataset_file)
                raise ValueError("All requests failed")

            logger.info("Finalizing writer")

            writer.finalize()

        output_dataset = Dataset.from_file(dataset_file)

        return output_dataset


def parse_response_message(
    response_message: str, response_format: Optional[BaseModel]
) -> tuple[Optional[dict | str], Optional[list[str]]]:
    response_errors = None
    if response_format:
        try:
            response_message = json.loads(response_message)
        except json.JSONDecodeError:
            logger.warning(
                f"Failed to parse response as JSON: {response_message}, skipping this response."
            )
            response_message = None
            response_errors = [
                f"Failed to parse response as JSON: {response_message}"
            ]
    return response_message, response_errors
